import json
import logging
import operator
from datetime import datetime
from functools import reduce
from pathlib import Path
from typing import Dict, List, Any, Mapping

import boto3
from botocore.client import BaseClient
from omegaconf import OmegaConf, dictconfig


# TODO: check and setup output type for all SM dependent method
# Sagemaker dependent methods

def get_pipeline_config(pipeline_module, config_type: str, pipeline_role: str, args: List) -> dictconfig:
    """
    Read pipeline config
    :param pipeline_module: python module which includes pipeline
    :param config_type: config filename
    :param pipeline_role: IAM role
    :param args: runtime pipeline arguments
    :return: OmegaConf Dict Config
    """
    default_conf_path = f'{Path(pipeline_module.__file__).parent}/{config_type}.yml'
    default_conf = OmegaConf.load(default_conf_path)
    arg_conf = OmegaConf.create({'pipeline': {'role': pipeline_role}})
    override_arg_conf = OmegaConf.from_dotlist(args)
    return OmegaConf.merge(default_conf, arg_conf, override_arg_conf)


def get_output_destination(sagemaker_client: BaseClient, processing_job_arn: str, output_name: str):
    pj_output_dict = _list_to_dict(
        sagemaker_client.describe_processing_job(
            ProcessingJobName=get_job_name(processing_job_arn)
        )['ProcessingOutputConfig']['Outputs'],
        'OutputName'
    )
    return pj_output_dict[output_name]['S3Output']['S3Uri']


def get_model_location(sagemaker_client: BaseClient, model_name: str):
    model_desc = sagemaker_client.describe_model(ModelName=model_name)
    return model_desc['PrimaryContainer']['ModelDataUrl']


def get_approved_package(sagemaker_client: BaseClient, model_package_group_name: str):
    response = sagemaker_client.list_model_packages(
        ModelApprovalStatus='Approved',
        ModelPackageGroupName=model_package_group_name,
        SortBy='CreationTime',
        SortOrder='Descending',
        MaxResults=1,
    )

    model_packages = response['ModelPackageSummaryList']
    if len(model_packages) == 0:
        raise ValueError(f"No approved ModelPackage found for ModelPackageGroup: {model_package_group_name}")

    return model_packages[0]


def load_json_from_s3(s3_uri: str) -> Dict:
    s3_client = boto3.client('s3')
    bucket, key = s3_uri.replace("s3://", "").split("/", 1)
    s3_response_object = s3_client.get_object(Bucket=bucket, Key=key)

    return json.loads(s3_response_object['Body'].read().decode('utf-8'))


def create_model_from_model_package(
        sagemaker_client: BaseClient,
        model_name: str,
        model_package_arn: str,
        execution_role: str,
        tags: List[Dict[str, str]],
) -> str:
    response = sagemaker_client.create_model(
        ModelName=model_name,
        Containers=[{'ModelPackageName': model_package_arn}],
        ExecutionRoleArn=execution_role,
        Tags=tags,
    )
    return response['ModelArn']


# Common methods

def get_datetime_str(date_time: datetime) -> str:
    return date_time.strftime('%Y-%m-%d-%H-%M-%S')


def convert_param_dict_to_key_value_list(arg_dict: Dict[str, str]) -> List[Dict[str, str]]:
    """
    Convert python dict to Sagemaker SDK resource tags structure
    where dict key corresponds to "Key", dict value corresponds to "Value".
    :param arg_dict: key-value need to convert to AWS resource tags structure
    :return: list of tags in the following format: [ { "Key": "...", "Value": "..." }, ... ]
    """
    return [
        {'Key': k, 'Value': v}
        for k, v in arg_dict.items()
    ]


def get_value_from_dict(data_dict: Dict[str, Any], path: List[str]) -> Mapping:
    """
       Get necessary metric from evaluation.json
       :param data_dict: dictionary that contains metrics {"regression_metrics": {"mse":..}}
       :param path: path to metric ['regression_metrics','mse',…]
       :return: dictionary that contains target metric:
       for example {mse: 2.1} --from--> { "Key 01": { "Key 02:…{mse: 2.1}"}
    """
    return reduce(operator.getitem, path, data_dict)


def normalize_key(key: str) -> str:
    return key.replace('_', '').lower()


def get_model_name(model_arn: str) -> str:
    return model_arn[model_arn.rindex('/') + 1:]


def get_job_name(job_arn: str) -> str:
    return job_arn[job_arn.rindex('/') + 1:]


def _list_to_dict(arg_list: List[Dict[str, Any]], dict_key_attr: str) -> Dict[Any, Dict]:
    return {
        o[dict_key_attr]: o
        for o in arg_list
    }


def ensure_min_length(argument: str, min_length: int) -> str:
    if len(argument) < min_length:
        return f"{argument}_ {'0' * (min_length - len(argument))}"
    else:
        return argument


def normalize_pipeline_name(name: str, name_max_len: int = 82) -> str:
    name_len = len(name)
    if name_len > name_max_len:
        logging.getLogger(__name__).info(
            'Provided pipeline name "%s" is too long (%d symbols, max allowed: 82). It will be truncated.',
            name,
            name_len
        )
        name = name[:name_max_len]
    return name


def _generate_data_capture_config(
        s3_destination_prefix: str,
        sampling_percentage: int = 100,
) -> Dict[str, Any]:
    return {
        'EnableCapture': True,
        'InitialSamplingPercentage': sampling_percentage,
        'DestinationS3Uri': s3_destination_prefix,
        'CaptureOptions': [{'CaptureMode': 'Input'}, {'CaptureMode': 'Output'}],  # both by default
        'CaptureContentTypeHeader': {'CsvContentTypes': ['text/csv']},
    }



import json
import logging
from datetime import datetime
from importlib import import_module
from typing import Dict, Optional, NoReturn, Any

import boto3
from omegaconf import OmegaConf
from sagemaker import (
    Predictor,
    ModelPackage,
    Session
)
from sagemaker.model_monitor import DataCaptureConfig

from mlops_utilities import helpers

logger = logging.getLogger(__name__)


def upsert_pipeline(
        pipeline_module: str,
        pipeline_package: str,
        pipeline_name: str,
        config_type: str,
        role: str,
        pipeline_tags: Optional[Dict[str, str]] = None,
        dryrun: bool = False,
        *args,
) -> NoReturn:
    """
    Performs Sagemaker pipeline creating or updating.

    Example:
    >>> upsert_pipeline('training_pipeline', 'src', 'a_cool_pipeline_name', 'training.defaults', 'role-arn', ...)

    First two arguments is set so to follow (folder) package structure described below
    when the function is invoked from the root dir:
        root
        |
        |-- src
        |   |-- training_pipeline.py
        |   |-- ...
        |   `-- ...
        |
        ...

    :param config_type: name of the pipeline yml file with configurations, <training_pipeline>.<config_type>
    :param role: your IAM role
    :param pipeline_module: a "module path" within the 'pipeline_package' (relative to the 'pipeline_package' root)
    :param pipeline_package: a package where 'pipeline_module' is defined
    :param pipeline_name: the name of the pipeline
    :param pipeline_tags: {"<key>": "<value>", ...} dict to be set as SM pipeline resource tags
    :param dryrun: whether to skip actual pipeline upsert or not
    :param args: extra configuration to pass to pipeline building;
        must follow dot-notation (https://omegaconf.readthedocs.io/en/2.0_branch/usage.html#from-a-dot-list)
    """
    pipeline_module = import_module(f'{pipeline_module}.{pipeline_package}')
    result_conf = helpers.get_pipeline_config(pipeline_module, config_type, role, list(args))

    if logger.isEnabledFor(logging.INFO):
        logger.info('Result config:\n%s', OmegaConf.to_yaml(result_conf, resolve=True))
    sm_session = Session(default_bucket=OmegaConf.select(result_conf, 'pipeline.default_bucket', default=None))

    pipeline_name = helpers.normalize_pipeline_name(pipeline_name)
    pipeline_object = pipeline_module.get_pipeline(sm_session, pipeline_name, result_conf)
    if logger.isEnabledFor(logging.INFO):
        logger.info("Pipeline definition:\n%s",
                    json.dumps(
                        json.loads(pipeline_object.definition()),
                        indent=2
                    ))

    if not dryrun:
        if pipeline_tags is not None:
            pipeline_tags = helpers.convert_param_dict_to_key_value_list(pipeline_tags)
        pipeline_object.upsert(result_conf.pipeline.role, tags=pipeline_tags)


def run_pipeline(
        pipeline_name: str,
        execution_name_prefix: str,
        pipeline_params: Dict[str, Any],
        dryrun=False) -> NoReturn:
    """
    Performs Sagemaker pipeline running.
    !!! This pipeline should be created and should be uploaded to Sagemaaker.

    Example:
    >>> run_pipeline('a_cool_pipeline_name', 'training_exec')

    :param pipeline_name: uploaded Sagemaker pipeline name
    :param execution_name_prefix: prefix for pipeline running job
    :param dryrun: should be run in test mode without real execution. If true then the method returns only arguments
    :param pipeline_params: additional parameters for pipeline
    """
    sagemaker_client = boto3.client('sagemaker')  # Can not be cut off because it could not be presented as string
    now = datetime.today()
    now_str = helpers.get_datetime_str(now)
    pipe_exec_name = f'{execution_name_prefix}-{now_str}'
    start_pipe_args = {
        'PipelineName': pipeline_name,
        'PipelineExecutionDisplayName': pipe_exec_name,
        'PipelineParameters': [
            {
                'Name': k,
                'Value': str(v)
            }
            for k, v in pipeline_params.items()
        ],
        'ClientRequestToken': helpers.ensure_min_length(pipe_exec_name, 32)
    }
    if dryrun:
        return start_pipe_args
    else:
        return sagemaker_client.start_pipeline_execution(**start_pipe_args)


def deploy_model(
        sagemaker_session: Session,
        model_package_group_name: str,
        instance_type: str,
        instance_count: int,
        endpoint_name: str,
        data_capture_s3_uri: str,
        role: str) -> NoReturn:
    """
    Method deploys model to Sagemaker
    :param sagemaker_session: Sagemaker session
    :param model_package_group_name: destination model package of deploying
    :param instance_type: instance types on which the model is deployed
    :param instance_count: amount of instances on which the model is deployed
    :param endpoint_name:
    :param data_capture_s3_uri: s3 bucket which accumulates data capture
    :param role: execution IAM role
    """
    instance_count = int(instance_count)

    sagemaker_client = sagemaker_session.sagemaker_client

    pck = helpers.get_approved_package(sagemaker_client, model_package_group_name)
    model_description = sagemaker_client.describe_model_package(ModelPackageName=pck["ModelPackageArn"])

    logger.info(f"EndpointName= {endpoint_name}")

    endpoints = sagemaker_client.list_endpoints(NameContains=endpoint_name)['Endpoints']

    data_capture_config = DataCaptureConfig(
        enable_capture=True, sampling_percentage=100, destination_s3_uri=data_capture_s3_uri
    )

    logger.info("Data capture enabled")

    if len(endpoints) > 0:
        logger.info("Update current endpoint")
        update_endpoint(sagemaker_client, instance_type, instance_count, endpoint_name, data_capture_config)
    else:
        logger.info("Create endpoint")

        model_package_arn = model_description["ModelPackageArn"]
        create_endpoint(model_package_arn, sagemaker_session, instance_count, instance_type, endpoint_name,
                        data_capture_config, role)


def compare_metrics(sagemaker_client,
                    endpoint_config_description: Dict[str, Any],
                    model_statistics_s3_uri: str,
                    metric: str,
                    dryrun: bool = False) -> bool:
    """
    This method compares metrics of old and new model versions
    :param sagemaker_client: boto3_session_client(sagemaker)
    :param endpoint_config_description: endpoint configuration
    :param model_statistics_s3_uri: s3 bucket which contains evaluation metrics
    :param metric: path to metric in json file, example: 'regression_metrics/mse/value'
    :param dryrun: is True in case of test
    :return: result of metric comparison
    """
    model_deployed_description = sagemaker_client.describe_model(
        ModelName=endpoint_config_description["ProductionVariants"][0]["ModelName"])
    model_deployed_description = sagemaker_client.describe_model_package(
        ModelPackageName=model_deployed_description["Containers"][0]["ModelPackageName"])
    test_json_metrics = {"regression_metrics": {"mse": {"value": 4}}}
    new_model_metrics = test_json_metrics if dryrun else helpers.load_json_from_s3(model_statistics_s3_uri)
    old_model_metrics = test_json_metrics if dryrun else helpers.load_json_from_s3(
        model_deployed_description["ModelMetrics"]["ModelQuality"]["Statistics"]["S3Uri"])
    metric_path = metric.split('/')

    new_metric = helpers.get_value_from_dict(new_model_metrics, metric_path[:-1])[metric_path[-1]]
    old_metric = helpers.get_value_from_dict(old_model_metrics, metric_path[:-1])[metric_path[-1]]
    return new_metric >= old_metric


def update_endpoint(sagemaker_client,
                    instance_type: str,
                    instance_count: int,
                    endpoint_name: str,
                    data_capture_config: DataCaptureConfig,
                    model_statistics_s3_uri: str = None,
                    metric: str = None,
                    dryrun: bool = False) -> NoReturn:
    """
    Updating Sagemaker endpoint
    :param sagemaker_client: boto3_session_client(sagemaker)
    :param endpoint_name:
    :param data_capture_config: config for inference data capture
    :param model_statistics_s3_uri: s3 bucket which contains evaluation metrics
    :param metric: path to metric value in `model_statistics_s3_uri`
    :param dryrun: is 'True' in a case of testing
    :return:
    """
    endpoint_config_description = sagemaker_client.describe_endpoint_config(EndpointConfigName=endpoint_name)
    model_name = endpoint_config_description["ProductionVariants"][0]["ModelName"] if dryrun is False else 'model'
    require_update = metric is None or (metric is not None and compare_metrics(sagemaker_client,
                                                                               endpoint_config_description,
                                                                               model_statistics_s3_uri,
                                                                               metric))

    if require_update:
        predictor = Predictor(endpoint_name=endpoint_name)
        predictor.update_endpoint(initial_instance_count=instance_count,
                                  instance_type=instance_type,
                                  model_name=model_name)
        predictor.update_data_capture_config(data_capture_config)
    else:
        logger.info(
            "Current endpoint is not updated because the new model have worse quality than current deployed model"
        )


def create_endpoint(model_package_arn: str,
                    sagemaker_session: Session,
                    instance_count: int,
                    instance_type: str,
                    endpoint_name: str,
                    data_capture_config: DataCaptureConfig,
                    role: str) -> NoReturn:
    """
    It executes endpoint creation into Sagemaker
    :param model_package_arn: model package descriptor
    :param sagemaker_session: Sagemaker session
    :param instance_count: amount of instances on which the model is deployed
    :param instance_type: instance types on which the model is deployed
    :param endpoint_name: endpoint name as string
    :param data_capture_config: config for inference data capture
    :param role: execution IAM role
    :return:
    """
    model = ModelPackage(
        role=role, model_package_arn=model_package_arn, sagemaker_session=sagemaker_session
    )

    model.deploy(initial_instance_count=instance_count, instance_type=instance_type, endpoint_name=endpoint_name,
                 data_capture_config=data_capture_config)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mlops_utilities']

package_data = \
{'': ['*']}

install_requires = \
['boto3>=1.26,<1.27',
 'omegaconf>=2.2,<2.3',
 'pytest==7.2.0',
 'sagemaker>=2.112,<2.113']

setup_kwargs = {
    'name': 'mlops-utilities',
    'version': '0.2.0',
    'description': '',
    'long_description': "# MLOps Utilities\n\nThis library provides implementation for a few high level operations most commonly occuring in any MLOps workflows built in AWS:\n- Upsert pipeline\n- Run pipeline\n- Deploy model\n    - Create endpoint\n    - Update endpoint\n\n### User guide\n#### Installation\n`pip3 install mlops-utilities`\n#### Project structure conventions\nTODO\n#### Actions\nTODO\n```\nfrom mlops_utilities.actions import run_pipeline\n    \nrun_pipeline(\n  pipeline_name='test_pipeline',\n  execution_name_prefix='test_pipeline',\n  dryrun=True,\n  pipeline_params={}\n)\n```\n",
    'author': 'Provectus Team',
    'author_email': 'mlops@provectus.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8',
}


setup(**setup_kwargs)
